@@ -39,6 +39,8 @@ import Dhall.Core
     , URL(..)
     , Var(..)
     )
+
+import Data.ByteArray.Encoding (Base(..))
 import Data.Foldable (toList)
 import Data.List.NonEmpty (NonEmpty(..))
 import Data.Monoid ((<>))
@@ -48,19 +50,23 @@ import Prelude hiding (exponent)
 import GHC.Float (double2Float, float2Double)
 import Codec.CBOR.Magic (floatToWord16, wordToFloat16)
 
+import qualified Crypto.Hash
+import qualified Data.ByteArray.Encoding
+import qualified Data.ByteString
 import qualified Data.Sequence
 import qualified Data.Text
+import qualified Data.Text.Encoding
 import qualified Dhall.Map
 import qualified Dhall.Set
 import qualified Options.Applicative
 
 -- | Supported version strings
 data StandardVersion
-    = V_4_0_0
-    -- ^ Version "4.0.0"
+    = V_5_0_0
+    -- ^ Version "5.0.0"
 
 defaultStandardVersion :: StandardVersion
-defaultStandardVersion = V_4_0_0
+defaultStandardVersion = V_5_0_0
 
 parseStandardVersion :: Parser StandardVersion
 parseStandardVersion =
@@ -74,7 +80,7 @@ parseStandardVersion =
     readVersion = do
         string <- Options.Applicative.str
         case string :: Text of
-            "4.0.0" -> return V_4_0_0
+            "5.0.0" -> return V_5_0_0
             _       -> fail "Unsupported version"
 
 {-| Convert a function applied to multiple arguments to the base function and
@@ -386,13 +392,21 @@ importToTerm import_ =
     case importType of
         Remote (URL { scheme = scheme₀, ..}) ->
             TList
-                (   [ TInt 24, TInt scheme₁, TString authority ]
+                (   prefix
+                ++  [ TInt scheme₁, using, TString authority ]
                 ++  map TString (reverse components)
                 ++  [ TString file ]
                 ++  (case query    of Nothing -> [ TNull ]; Just q -> [ TString q ])
                 ++  (case fragment of Nothing -> [ TNull ]; Just f -> [ TString f ])
                 )
           where
+            using = case headers of
+                Nothing ->
+                    TNull
+                Just h ->
+                    importToTerm
+                        (Import { importHashed = h, importMode = Code })
+
             scheme₁ = case scheme₀ of
                 HTTP  -> 0
                 HTTPS -> 1
@@ -402,7 +416,8 @@ importToTerm import_ =
 
         Local prefix₀ path ->
                 TList
-                    (   [ TInt 24, TInt prefix₁ ]
+                    (   prefix
+                    ++  [ TInt prefix₁ ]
                     ++  map TString components₁
                     ++  [ TString file ]
                     )
@@ -411,18 +426,31 @@ importToTerm import_ =
 
             Directory {..} = directory
 
-            (prefix₁, components₁) = case (prefix₀, reverse components) of
-                (Absolute, rest       ) -> (2, rest)
-                (Here    , ".." : rest) -> (4, rest)
-                (Here    , rest       ) -> (3, rest)
-                (Home    , rest       ) -> (5, rest)
+            prefix₁ = case prefix₀ of
+              Absolute -> 2
+              Here     -> 3
+              Parent   -> 4
+              Home     -> 5
+
+            components₁ = reverse components
 
         Env x ->
-            TList [ TInt 24, TInt 6, TString x ]
+            TList (prefix ++ [ TInt 6, TString x ])
 
         Missing ->
-            TList [ TInt 24, TInt 7 ]
+            TList (prefix ++ [ TInt 7 ])
   where
+    prefix = [ TInt 24, h, m ]
+      where
+        h = case hash of
+            Nothing ->
+                TNull
+            Just digest ->
+                TList
+                    [ TString "sha256", TString (Data.Text.pack (show digest)) ]
+
+        m = TInt (case importMode of Code -> 0; RawText -> 1)
+
     Import {..} = import_
 
     ImportHashed {..} = importHashed
@@ -661,7 +689,28 @@ decode (TList (TInt 18 : xs)) = do
     (xys, z) <- process xs
 
     return (TextLit (Chunks xys z))
-decode (TList (TInt 24 : TInt n : xs)) = do
+decode (TList (TInt 24 : h : TInt mode : TInt n : xs)) = do
+    hash <- case h of
+        TNull -> do
+            return Nothing
+
+        TList [ TString "sha256", TString base16Text ] -> do
+            let base16Bytes = Data.Text.Encoding.encodeUtf8 base16Text
+            digestBytes <- case Data.ByteArray.Encoding.convertFromBase Base16 base16Bytes of
+                Left  _           -> empty
+                Right digestBytes -> return (digestBytes :: Data.ByteString.ByteString)
+
+            digest <- Crypto.Hash.digestFromByteString digestBytes
+            return (Just digest)
+
+        _ -> do
+            empty
+
+    importMode <- case mode of
+        0 -> return Code
+        1 -> return RawText
+        _ -> empty
+
     let remote scheme = do
             let process [ TString file, q, f ] = do
                     query <- case q of
@@ -679,16 +728,21 @@ decode (TList (TInt 24 : TInt n : xs)) = do
                 process _ = do
                     empty
 
-            (authority, paths, file, query, fragment) <- case xs of
-                TString authority : ys -> do
+            (headers, authority, paths, file, query, fragment) <- case xs of
+                headers₀ : TString authority : ys -> do
+                    headers₁ <- case headers₀ of
+                        TNull -> return Nothing
+                        _     -> do
+                            Embed (Import { importHashed = headers }) <- decode headers₀
+                            return (Just headers)
                     (paths, file, query, fragment) <- process ys
-                    return (authority, paths, file, query, fragment)
-                _                      -> empty
+                    return (headers₁, authority, paths, file, query, fragment)
+                _ -> do
+                    empty
 
             let components = reverse paths
             let directory  = Directory {..}
             let path       = File {..}
-            let headers    = Nothing
 
             return (Remote (URL {..}))
 
@@ -703,11 +757,7 @@ decode (TList (TInt 24 : TInt n : xs)) = do
 
             (paths, file) <- process xs
 
-            let finalPaths = case n of
-                    4 -> ".." : paths
-                    _ -> paths
-
-            let components = reverse finalPaths
+            let components = reverse paths
             let directory  = Directory {..}
 
             return (Local prefix (File {..}))
@@ -724,15 +774,14 @@ decode (TList (TInt 24 : TInt n : xs)) = do
         1 -> remote HTTPS
         2 -> local Absolute
         3 -> local Here
-        4 -> local Here
+        4 -> local Parent
         5 -> local Home
         6 -> env
         7 -> missing
         _ -> empty
 
-    let hash         = Nothing
     let importHashed = ImportHashed {..}
-    let importMode   = Code
+
     return (Embed (Import {..}))
 decode (TList (TInt 25 : xs)) = do
     let process (TString x : _A₁ : a₁ : ls₁) = do
@@ -778,7 +827,7 @@ decodeWithVersion term = do
             fail ("Cannot decode the version from this decoded CBOR expression: " <> show term)
 
     case version of
-        "4.0.0" -> do
+        "5.0.0" -> do
             return ()
         _ -> do
             fail ("This decoded version is not supported: " <> Data.Text.unpack version)
@@ -791,8 +840,8 @@ decodeWithVersion term = do
 
 -- | Encode a Dhall expression using the specified `Version`
 encodeWithVersion :: StandardVersion -> Expr s Import -> Term
-encodeWithVersion V_4_0_0 expression =
-    TList [ TString "4.0.0", encode expression ]
+encodeWithVersion V_5_0_0 expression =
+    TList [ TString "5.0.0", encode expression ]
 
 data DecodingFailure
     = CannotDecodeVersionString Term
