@@ -157,6 +157,8 @@ data FilePrefix
     -- ^ Absolute path
     | Here
     -- ^ Path relative to @.@
+    | Parent
+    -- ^ Path relative to @..@
     | Home
     -- ^ Path relative to @~@
     deriving (Eq, Generic, Ord, Show)
@@ -164,6 +166,7 @@ data FilePrefix
 instance Pretty FilePrefix where
     pretty Absolute = ""
     pretty Here     = "."
+    pretty Parent   = ".."
     pretty Home     = "~"
 
 data Scheme = HTTP | HTTPS deriving (Eq, Generic, Ord, Show)
@@ -188,12 +191,21 @@ data ImportType
     | Missing
     deriving (Eq, Generic, Ord, Show)
 
+parent :: File
+parent = File { directory = Directory { components = [ ".." ] }, file = "" }
+
 instance Semigroup ImportType where
     Local prefix file₀ <> Local Here file₁ = Local prefix (file₀ <> file₁)
 
     Remote (URL { path = path₀, ..}) <> Local Here path₁ =
         Remote (URL { path = path₀ <> path₁, ..})
 
+    Local prefix file₀ <> Local Parent file₁ =
+        Local prefix (file₀ <> parent <> file₁)
+
+    Remote (URL { path = path₀, .. }) <> Local Parent path₁ =
+        Remote (URL { path = path₀ <> parent <> path₁, .. })
+
     import₀ <> Remote (URL { headers = headers₀, .. }) =
         Remote (URL { headers = headers₁, .. })
       where
@@ -1575,14 +1587,16 @@ normalizeWithM ctx e0 = loop (denote e0)
           Just e1 -> loop e1
           Nothing -> do
               f' <- loop f
+              a' <- loop a
               case f' of
-                Lam x _A b -> loop b''
-                  where
-                    a'  = shift   1  (V x 0) a
-                    b'  = subst (V x 0) a' b
-                    b'' = shift (-1) (V x 0) b'
+                Lam x _A b₀ -> do
+
+                    let a₂ = shift 1 (V x 0) a'
+                    let b₁ = subst (V x 0) a₂ b₀
+                    let b₂ = shift (-1) (V x 0) b₁
+
+                    loop b₂
                 _ -> do
-                  a' <- loop a
                   case App f' a' of
                     -- build/fold fusion for `List`
                     App (App ListBuild _) (App (App ListFold _) e') -> loop e'
@@ -1597,8 +1611,8 @@ normalizeWithM ctx e0 = loop (denote e0)
                       t' <- loop t
                       if boundedType t' then strict else lazy
                       where
-                        strict =       strictLoop n0
-                        lazy   = loop (  lazyLoop n0)
+                        strict =       strictLoop (fromIntegral n0 :: Integer)
+                        lazy   = loop (  lazyLoop (fromIntegral n0 :: Integer))
 
                         strictLoop !0 = loop zero
                         strictLoop !n = App succ' <$> strictLoop (n - 1) >>= loop
@@ -1702,15 +1716,18 @@ normalizeWithM ctx e0 = loop (denote e0)
                         case res2 of
                             Nothing -> pure (App f' a')
                             Just app' -> loop app'
-    Let (Binding x _ a₀ :| ls₀) b₀ -> loop b₂
-      where
-        rest = case ls₀ of
-            []       -> b₀
-            l₁ : ls₁ -> Let (l₁ :| ls₁) b₀
+    Let (Binding x _ a₀ :| ls₀) b₀ -> do
+        a₁ <- loop a₀
 
-        a₁ = shift 1 (V x 0) a₀
-        b₁ = subst (V x 0) a₁ rest
-        b₂ = shift (-1) (V x 0) b₁
+        rest <- case ls₀ of
+                []       -> loop b₀
+                l₁ : ls₁ -> loop (Let (l₁ :| ls₁) b₀)
+
+        let a₂ = shift 1 (V x 0) a₁
+        let b₁ = subst (V x 0) a₂ rest
+        let b₂ = shift (-1) (V x 0) b₁
+
+        loop b₂
     Annot x _ -> loop x
     Bool -> pure Bool
     BoolLit b -> pure (BoolLit b)
@@ -1897,14 +1914,8 @@ normalizeWithM ctx e0 = loop (denote e0)
     Constructors t   -> do
         t' <- loop t
         case t' of
-            Union kts -> pure (RecordLit kvs)
-              where
-                kvs = Dhall.Map.mapWithKey adapt kts
-
-                adapt k t_ = Lam k t_ (UnionLit k (Var (V k 0)) rest)
-                  where
-                    rest = Dhall.Map.delete k kts
-            _ -> pure (Constructors t')
+            u@(Union _) -> pure u
+            _           -> pure $ Constructors t'
     Field r x        -> do
         r' <- loop r
         case r' of
